/**********************************************************************
		NOTEMIDI - module for serial port MIDI devices

		(C) 2001 by Michael Minn - michael@michaelminn.com

	printk messages are written to /var/log/messages

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

***************************************************************************/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/soundcard.h>
#include <linux/fs.h>
#include <linux/tty.h>
#include <linux/init.h>
#include <linux/timer.h>
#include <asm/uaccess.h>

#include <linux/delay.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
#include <linux/wait.h>
#include <linux/compatmac.h>
#include <linux/init.h>
#include <linux/sched.h>
#endif

#define NOTEMIDI_SERIAL_PORT_DEVICE_NAME "/dev/ttyS0"

#define NOTEMIDI_MAJOR_NUMBER 80
#define NOTEMIDI_SEQUENCER_MINOR_NUMBER 0
#define NOTEMIDI_MIDI_MINOR_NUMBER 1
#define NOTEMIDI_IS_MS124W_PROTOCOL 0

MODULE_AUTHOR("Michael Minn <michael@michaelminn.com>");
MODULE_DESCRIPTION("NOTEMIDI serial port MIDI driver");

static struct file *notemidi_serial_port_file = NULL;
static struct tty_struct *notemidi_serial_port_tty = NULL;
static int notemidi_open_time = 0;
static int notemidi_timer_start = 0; // in jiffies - i.e. 100 ticks per second
static int notemidi_timer_base = 96; // timer ticks per beat
static int notemidi_timer_tempo = 120; // beats per second
static int notemidi_echo_sync_time = 0;

static unsigned char notemidi_outbuffer[2048];
static int notemidi_outbuffer_pointer = 0;
static int notemidi_outbuffer_length = 0;

static struct timer_list notemidi_buffer_timer;
static wait_queue_head_t notemidi_buffer_wait_queue;

static int notemidi_echo_input = 1;

/**********************************************************************************
*   Serial Port write/read
*
*	For info on the TTY serial port drivers, see
*	/usr/include/linux/tty_driver.h
*	/usr/src/linux/drivers/char/serial.c
*
*	Because serial port handling is extremely complex, NOTEMIDI uses the kernel
*	serial port functions.  Serial port input calls the line discipline functions
*	directly. This is not portable.  Reads from the line discipline do not block, so
*	schedule_timeout is used to poll the read function.
*
*	For MS-124 write protocol - one Burst Byte preceeds each data byte
*	lower 4 bits always 8
*	upper 4 bits correspond to output port
*		0x18 = port 1
*		0x28 = port 2
*		0x48 = port 3
*		0x88 = port 4
*		0xf8 = all four ports simultaneously
**********************************************************************************/

static struct tty_struct *notemidi_open_serial_port(char *serial_port_device_name)
{
	struct tty_struct *tty = NULL;
	struct termios old_termios;

	if (IS_ERR(notemidi_serial_port_file = filp_open(serial_port_device_name, O_WRONLY, 0)))
		printk(KERN_INFO "NOTEMIDI: Failure %d opening serial port %s\n", (int) PTR_ERR(notemidi_serial_port_file), serial_port_device_name);

	else if ((tty = (struct tty_struct *)notemidi_serial_port_file->private_data) == NULL)
		printk(KERN_INFO "NOTEMIDI: NULL file private data TTY pointer\n");

	else if (tty->driver.set_termios == NULL)
		printk(KERN_INFO "NOTEMIDI: NULL file set_termios function\n");

	else if (tty->termios == NULL)
		printk(KERN_INFO "NOTEMIDI: NULL file termios structure\n");

	else if (tty->ldisc.flush_buffer == NULL)
		printk(KERN_ERR "NOTEMIDI: NULL TTY ldisc.flush_buffer() function\n");

	else if (tty->driver.write == NULL)
		printk(KERN_ERR "NOTEMIDI: NULL TTY driver write function\n");

	else if (tty->ldisc.read == NULL)
		printk(KERN_ERR "NOTEMIDI: NULL TTY ldisc.read() function\n");

	else
	{
		old_termios = *tty->termios;
		tty->termios->c_lflag = 0;
		tty->termios->c_lflag = NOFLSH;
		tty->termios->c_iflag = IGNBRK | IGNPAR;
		tty->termios->c_oflag = 0;
		tty->termios->c_cflag =  B38400 | CREAD | CSIZE | CS8 | CLOCAL | HUPCL;
		tty->termios->c_cc[VEOL] = 0; /* '\r'; */
		tty->termios->c_cc[VERASE] = 0;
		tty->termios->c_cc[VKILL] = 0;
		tty->termios->c_cc[VMIN] = 0;
		tty->termios->c_cc[VTIME] = 0;
		(*tty->driver.set_termios)(tty, &old_termios);
		return tty;
	}

	return 0;
}

static int notemidi_write_to_port(const char *data, size_t length)
{
	int scan;
	for (scan = 0; scan < length; ++scan)
	{
		int status = -EFAULT;
		char burst_byte = 0xf8;

		/* printk(KERN_INFO "NOTEMIDI: write %x\n", (unsigned char) data[scan]);  */

		if (NOTEMIDI_IS_MS124W_PROTOCOL && ((status = notemidi_serial_port_tty->driver.write(notemidi_serial_port_tty, 0, &burst_byte, 1)) < 1))
		{
			printk(KERN_ERR "NOTEMIDI: ms124w burst byte write failure: status %d\n", status);
			return -1;
		}

		else if ((status = notemidi_serial_port_tty->driver.write(notemidi_serial_port_tty, 0, &data[scan], 1)) < 1)
		{
			printk(KERN_ERR "NOTEMIDI: write failure: %d of %d bytes, status %d\n", scan, length, status);
			return -1;
		}
	}

	return length;
}



/**********************************************************************************
*   Kernel timed serial port input / output functions
*	/dev/sequencer requires MIDI input and output to be handled asynchronously
**********************************************************************************/

static void notemidi_write_outbuffer(unsigned long argument)
{
	unsigned char output[8];

	/* printk(KERN_INFO "NOTEMIDI: writing outbuffer: %d - %d\n", notemidi_outbuffer_pointer, notemidi_outbuffer_length); */

	for (; notemidi_outbuffer_pointer < notemidi_outbuffer_length; notemidi_outbuffer_pointer += 8)
	{
		/* printk(KERN_INFO "NOTEMIDI: sequencer event at %d: %x %x %x %x\n", notemidi_outbuffer_pointer, notemidi_outbuffer[notemidi_outbuffer_pointer], notemidi_outbuffer[notemidi_outbuffer_pointer + 1], notemidi_outbuffer[notemidi_outbuffer_pointer + 2], notemidi_outbuffer[notemidi_outbuffer_pointer + 3]); */

		if (notemidi_outbuffer[notemidi_outbuffer_pointer] == SEQ_MIDIPUTC)
		{
			output[0] = notemidi_outbuffer[notemidi_outbuffer_pointer + 1];
			if (notemidi_write_to_port(output, 1) < 1)
				printk(KERN_ERR "NOTEMIDI: Failure writing single MIDI byte\n");
			notemidi_outbuffer_pointer -= 4;
		}
		else if (notemidi_outbuffer[notemidi_outbuffer_pointer] == EV_CHN_VOICE)
		{
			output[0] = notemidi_outbuffer[notemidi_outbuffer_pointer + 2] | (notemidi_outbuffer[notemidi_outbuffer_pointer + 3] & 0x0f);
			output[1] = notemidi_outbuffer[notemidi_outbuffer_pointer + 4];
			output[2] = notemidi_outbuffer[notemidi_outbuffer_pointer + 5];
			if (notemidi_write_to_port(output, 3) < 3)
				printk(KERN_ERR "NOTEMIDI: Failure writing voice MIDI message\n");
		}
		else if ((notemidi_outbuffer[notemidi_outbuffer_pointer] == EV_CHN_COMMON) && (notemidi_outbuffer[notemidi_outbuffer_pointer + 2] == MIDI_CHN_PRESSURE))
		{
			output[0] = notemidi_outbuffer[notemidi_outbuffer_pointer + 2] | (notemidi_outbuffer[notemidi_outbuffer_pointer + 3] & 0x0f);
			output[1] = notemidi_outbuffer[notemidi_outbuffer_pointer + 4];
			if (notemidi_write_to_port(output, 2) < 2)
				printk(KERN_ERR "NOTEMIDI: Failure writing pressure message\n");
		}
		else if ((notemidi_outbuffer[notemidi_outbuffer_pointer] == EV_CHN_COMMON) && (notemidi_outbuffer[notemidi_outbuffer_pointer + 2] == MIDI_PGM_CHANGE))
		{
			output[0] = notemidi_outbuffer[notemidi_outbuffer_pointer + 2] | (notemidi_outbuffer[notemidi_outbuffer_pointer + 3] & 0x0f);
			output[1] = notemidi_outbuffer[notemidi_outbuffer_pointer + 4];
			if (notemidi_write_to_port(output, 2) < 2)
				printk(KERN_ERR "NOTEMIDI: Failure writing program message\n");
		}
		else if ((notemidi_outbuffer[notemidi_outbuffer_pointer] == EV_CHN_COMMON) && (notemidi_outbuffer[notemidi_outbuffer_pointer + 2] == MIDI_CTL_CHANGE))
		{
			output[0] = notemidi_outbuffer[notemidi_outbuffer_pointer + 2] | (notemidi_outbuffer[notemidi_outbuffer_pointer + 3] & 0x0f);
			output[1] = notemidi_outbuffer[notemidi_outbuffer_pointer + 4];
			output[2] = notemidi_outbuffer[notemidi_outbuffer_pointer + 6];
			if (notemidi_write_to_port(output, 3) < 3)
				printk(KERN_ERR "NOTEMIDI: Failure writing control message\n");
		}
		else if ((notemidi_outbuffer[notemidi_outbuffer_pointer] == EV_CHN_COMMON) && (notemidi_outbuffer[notemidi_outbuffer_pointer + 2] == MIDI_PITCH_BEND))
		{
			output[0] = notemidi_outbuffer[notemidi_outbuffer_pointer + 2] | (notemidi_outbuffer[notemidi_outbuffer_pointer + 3] & 0x0f);
			output[1] = notemidi_outbuffer[notemidi_outbuffer_pointer + 6];
			output[2] = notemidi_outbuffer[notemidi_outbuffer_pointer + 7];
			if (notemidi_write_to_port(output, 3) < 3)
				printk(KERN_ERR "NOTEMIDI: Failure writing bend message\n");
		}

		else if ((notemidi_outbuffer[notemidi_outbuffer_pointer] == EV_TIMING) && (notemidi_outbuffer[notemidi_outbuffer_pointer + 1] == TMR_START))
		{
			notemidi_timer_start = jiffies;
		}

		else if ((notemidi_outbuffer[notemidi_outbuffer_pointer] == EV_TIMING) && (notemidi_outbuffer[notemidi_outbuffer_pointer + 1] == TMR_STOP))
		{
			notemidi_timer_start = 0;
		}

		else if ((notemidi_outbuffer[notemidi_outbuffer_pointer] == EV_TIMING) && (notemidi_outbuffer[notemidi_outbuffer_pointer + 1] == TMR_CONTINUE))
		{
			notemidi_timer_start = jiffies;
		}

		else if ((notemidi_outbuffer[notemidi_outbuffer_pointer] == EV_TIMING) && (notemidi_outbuffer[notemidi_outbuffer_pointer + 1] == TMR_TEMPO))
		{
			notemidi_timer_tempo = *((unsigned int*)&notemidi_outbuffer[notemidi_outbuffer_pointer + 4]);
		}

		else if ((notemidi_outbuffer[notemidi_outbuffer_pointer] == EV_TIMING) && (notemidi_outbuffer[notemidi_outbuffer_pointer + 1] == TMR_TIMESIG))
		{
			/* doesn't do anything now, but accept it */
		}

		else if ((notemidi_outbuffer[notemidi_outbuffer_pointer] == EV_TIMING) && (notemidi_outbuffer[notemidi_outbuffer_pointer + 1] == TMR_ECHO))
		{
			notemidi_echo_sync_time = *((int*)&notemidi_outbuffer[notemidi_outbuffer_pointer + 4]);
		}

		else if ((notemidi_outbuffer[notemidi_outbuffer_pointer] == EV_TIMING) && (notemidi_outbuffer[notemidi_outbuffer_pointer + 1] == TMR_WAIT_REL))
		{
			int wait_ticks = *((unsigned int*)&notemidi_outbuffer[notemidi_outbuffer_pointer + 4]);
			double wait_jiffies = ((double)wait_ticks) / ((double)notemidi_timer_base) * 60.0 * ((double)HZ) / ((double)notemidi_timer_tempo);
			double absolute_jiffies = ((double)jiffies) + wait_jiffies - ((double)notemidi_timer_start);
			double absolute_ticks = absolute_jiffies * ((double)notemidi_timer_base) / 60.0 / ((double)HZ) * ((double)notemidi_timer_tempo);

			// Convert relative time messages to absolute time for easier handling
			notemidi_outbuffer[notemidi_outbuffer_pointer + 1] = TMR_WAIT_ABS;
			*((unsigned int*)&notemidi_outbuffer[notemidi_outbuffer_pointer + 4]) = ((int)absolute_ticks);

			// printk(KERN_INFO "NOTEMIDI: relative wait %d = absolute %e\n", wait_ticks, absolute_ticks);
			notemidi_outbuffer_pointer -= 8; // so this event is handled again by the loop
		}

		else if ((notemidi_outbuffer[notemidi_outbuffer_pointer] == EV_TIMING) && (notemidi_outbuffer[notemidi_outbuffer_pointer + 1] == TMR_WAIT_ABS))
		{
			int wait_ticks = *((unsigned int*)&notemidi_outbuffer[notemidi_outbuffer_pointer + 4]);
			double wait_jiffies = ((double)wait_ticks) / ((double)notemidi_timer_base) * 60.0 * ((double)HZ) / ((double)notemidi_timer_tempo);
			int expiration_jiffies = ((int)wait_jiffies) + notemidi_timer_start;

			/* printk(KERN_INFO "NOTEMIDI: absolute wait %d - expires %d, it is now %d\n", wait_ticks, expiration_jiffies, (int) jiffies); */

			if (notemidi_timer_start <= 0)
				printk(KERN_ERR "NOTEMIDI: wait with no timer started\n");
			else
			{
				init_timer(&notemidi_buffer_timer);
				notemidi_buffer_timer.expires = expiration_jiffies;
				notemidi_buffer_timer.data = 0;
				notemidi_buffer_timer.function = notemidi_write_outbuffer;
				add_timer(&notemidi_buffer_timer);
				notemidi_outbuffer_pointer += 8;
				return;
			}
		}

		else if (SYSEX_PATCH == *((short*)&notemidi_outbuffer[notemidi_outbuffer_pointer]))
		{
			struct sysex_info *sysex = (struct sysex_info*) &notemidi_outbuffer[notemidi_outbuffer_pointer];

			if ((sysex->len + notemidi_outbuffer_pointer + 8) > notemidi_outbuffer_length)
				sysex->len = notemidi_outbuffer_length - notemidi_outbuffer_pointer - 8;

			if (notemidi_write_to_port(&notemidi_outbuffer[notemidi_outbuffer_pointer + 8], sysex->len) < sysex->len)
				printk(KERN_ERR "NOTEMIDI: Failure writing sysex message\n");

			notemidi_outbuffer_pointer += sysex->len;
		}

		else
			printk(KERN_INFO "NOTEMIDI: unsupported sequencer event %x %x %x %x %x %x %x %x\n", notemidi_outbuffer[notemidi_outbuffer_pointer], notemidi_outbuffer[notemidi_outbuffer_pointer + 1], notemidi_outbuffer[notemidi_outbuffer_pointer + 2], notemidi_outbuffer[notemidi_outbuffer_pointer + 3], notemidi_outbuffer[notemidi_outbuffer_pointer + 4], notemidi_outbuffer[notemidi_outbuffer_pointer + 5], notemidi_outbuffer[notemidi_outbuffer_pointer + 6], notemidi_outbuffer[notemidi_outbuffer_pointer + 7]);
	}

	/* printk(KERN_INFO "NOTEMIDI: outbuffer flushed - indexes now %d - %d\n", notemidi_outbuffer_pointer, notemidi_outbuffer_length); */

	notemidi_outbuffer_pointer = 0;
	notemidi_outbuffer_length = 0;

	if (waitqueue_active(&notemidi_buffer_wait_queue))
		wake_up_interruptible(&notemidi_buffer_wait_queue);

	return;
}


/****************************************************************************
*	/dev/midiXX file operations
*
*	It is possible to override the standard OSS file_operations functions
*	with module supplied file_operations.  There should be no conflicts
*	with interrupts from /dev/sequencer which calls the OSS midi_operations.
*
*****************************************************************************/


static int notemidi_file_open (struct inode *inode, struct file *file)
{
	printk(KERN_INFO "NOTEMIDI: open minor number %d\n", MINOR(inode->i_rdev));

	if (notemidi_open_time > 0)
		return -EBUSY;

	file->private_data = (void*) MINOR(inode->i_rdev);
	notemidi_timer_base = 50;
	notemidi_timer_tempo = 120; // default is 100 ticks/second - same as Linux jiffies clock (i.e. HZ)
	notemidi_timer_start = 0;
	notemidi_open_time = jiffies;

	notemidi_outbuffer_pointer = 0;
	notemidi_outbuffer_length = 0;
	init_timer(&notemidi_buffer_timer);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
	init_waitqueue(&notemidi_buffer_wait_queue);
#else
	init_waitqueue_head(&notemidi_buffer_wait_queue);
#endif
	MOD_INC_USE_COUNT;

	return 0;
}

static int notemidi_file_release (struct inode *inode, struct file *file)
{
	printk(KERN_INFO "NOTEMIDI: close minor number %d\n", MINOR(inode->i_rdev));

	if (waitqueue_active(&notemidi_buffer_wait_queue))
		wake_up_interruptible(&notemidi_buffer_wait_queue);

	del_timer(&notemidi_buffer_timer);

	notemidi_open_time = 0;
	notemidi_timer_start = 0;

	MOD_DEC_USE_COUNT;

	return 0;
}

static ssize_t notemidi_file_write (struct file *file, const char *data, size_t length, loff_t *offset)
{
	/* printk(KERN_INFO "NOTEMIDI: write %d bytes to minor number %d\n", length, (int) file->private_data); */

	if (((int) file->private_data) == NOTEMIDI_MIDI_MINOR_NUMBER)
	{
		copy_from_user(notemidi_outbuffer, data, length);
		return notemidi_write_to_port(notemidi_outbuffer, length);
	}

	if ((file->f_flags & O_NDELAY) && notemidi_outbuffer_length)
		return -EAGAIN;

	else if (notemidi_outbuffer_length)
		interruptible_sleep_on_timeout(&notemidi_buffer_wait_queue, 1000);

	if (notemidi_outbuffer_length)
	{
		printk(KERN_ERR "NOTEMIDI: Timeout while waiting for write buffer to flush\n");
		return -1;
	}

	/* printk(KERN_INFO "NOTEMIDI: copying %d bytes to buffer index %d\n", length, notemidi_outbuffer_pointer); */
	if (length > sizeof(notemidi_outbuffer))
		length = sizeof(notemidi_outbuffer);

	copy_from_user(notemidi_outbuffer, data, length);
	notemidi_outbuffer_length = length;
	notemidi_outbuffer_pointer = 0;

	init_timer(&notemidi_buffer_timer);
	notemidi_buffer_timer.expires = jiffies + 1;
	notemidi_buffer_timer.data = 0;
	notemidi_buffer_timer.function = notemidi_write_outbuffer;
	add_timer(&notemidi_buffer_timer);

	return length;
}

static ssize_t notemidi_file_read (struct file *file, char *data, size_t length, loff_t *offset)
{
	int scan = 0;
	static int previous_timer = 0;
	static unsigned char running_status = 0;
	int current_jiffies = jiffies - notemidi_timer_start;
	double real_timer = ((double)notemidi_timer_base) * ((double)notemidi_timer_tempo) * ((double)current_jiffies) / 60.0 / ((double)HZ);
	int current_timer = (int) real_timer;

	if (((int) file->private_data) == NOTEMIDI_MIDI_MINOR_NUMBER)
		return notemidi_serial_port_tty->ldisc.read(notemidi_serial_port_tty, notemidi_serial_port_file, data, length);

	if (notemidi_echo_sync_time)
	{
		// printk(KERN_INFO "NOTEMIDI: sync echo read\n");
		put_user((unsigned char)EV_TIMING, &data[0]);
		put_user((unsigned char)TMR_ECHO, &data[1]);
		put_user((unsigned char)0, &data[2]);
		put_user((unsigned char)0, &data[3]);
		put_user(current_timer, (int*)&data[4]);
		notemidi_echo_sync_time = 0;
		scan += 8;
	}

	else if ((notemidi_timer_start > 0) && (current_timer != previous_timer))
	{
		put_user((unsigned char)EV_TIMING, &data[0]);
		put_user((unsigned char)TMR_WAIT_ABS, &data[1]);
		put_user((unsigned char)0, &data[2]);
		put_user((unsigned char)0, &data[3]);
		put_user(current_timer, (int*)&data[4]);
		previous_timer = current_timer;
		scan += 8;
	}

	while (notemidi_serial_port_tty->ldisc.read(notemidi_serial_port_tty, notemidi_serial_port_file, &data[scan], 1) == 1)
	{
		unsigned char status_byte = 0;
		unsigned char second_byte = 0;
		unsigned char third_byte = 0;
		int retry = 200;

		get_user(status_byte, &data[scan]);
		/* printk(KERN_INFO "NOTEMIDI: read status byte %d\n", status_byte); */

		if (status_byte < 0x70)
		{
			second_byte = status_byte;
			status_byte = running_status;
			/* printk(KERN_INFO "NOTEMIDI: running status %x %x\n", status_byte, second_byte); */
		}
		else if (status_byte != 0xfe) // ignore active sensing
		{
			running_status = status_byte;

			while (retry && (notemidi_serial_port_tty->ldisc.read(notemidi_serial_port_tty, notemidi_serial_port_file, &data[scan], 1) == 0))
				--retry;

			get_user(second_byte, &data[scan]);
			/* printk(KERN_INFO "NOTEMIDI: new status %x %x\n", status_byte, second_byte); */
		}

		if (retry)
			switch (status_byte & 0xf0)
			{
				case MIDI_PGM_CHANGE:
				case MIDI_CHN_PRESSURE:
					put_user((unsigned char) EV_CHN_COMMON, &data[scan]);
					put_user((unsigned char) 0, &data[scan + 1]);
					put_user((unsigned char) status_byte & 0xf0, &data[scan + 2]);
					put_user((unsigned char) status_byte & 0x0f, &data[scan + 3]);
					put_user((unsigned char) second_byte, &data[scan + 4]);
					scan += 8;

					if (notemidi_echo_input)
					{
						notemidi_write_to_port(&status_byte, 1);
						notemidi_write_to_port(&second_byte, 1);
					}
				break;

				case MIDI_CTL_CHANGE:
				case MIDI_PITCH_BEND:
					put_user((unsigned char) EV_CHN_COMMON, &data[scan]);
					put_user((unsigned char) 0, &data[scan + 1]);
					put_user((unsigned char) status_byte & 0xf0, &data[scan + 2]);
					put_user((unsigned char) status_byte & 0x0f, &data[scan + 3]);
					put_user((unsigned char) second_byte, &data[scan + 6]);

					/* printk(KERN_INFO "NOTEMIDI: 3 byte channel common msg\n"); */
					while (retry && (notemidi_serial_port_tty->ldisc.read(notemidi_serial_port_tty, notemidi_serial_port_file, &data[scan + 7], 1) == 0))
						--retry;

					if (notemidi_echo_input)
					{
						notemidi_write_to_port(&status_byte, 1);
						notemidi_write_to_port(&second_byte, 1);
						copy_from_user(&third_byte, &data[scan + 7], 7);
						notemidi_write_to_port(&third_byte, 1);
					}

					if (retry)
						scan += 8;
					break;

				case MIDI_NOTEOFF:
				case MIDI_NOTEON:
				case MIDI_KEY_PRESSURE:
					put_user((unsigned char) EV_CHN_VOICE, &data[scan]);
					put_user((unsigned char) 0, &data[scan + 1]);
					put_user((unsigned char) status_byte & 0xf0, &data[scan + 2]);
					put_user((unsigned char) status_byte & 0x0f, &data[scan + 3]);
					put_user((unsigned char) second_byte, &data[scan + 4]);

					/* printk(KERN_INFO "NOTEMIDI: voice msg\n"); */
					while (retry && (notemidi_serial_port_tty->ldisc.read(notemidi_serial_port_tty, notemidi_serial_port_file, &data[scan + 5], 1) == 0))
						--retry;

					if (notemidi_echo_input)
					{
						notemidi_write_to_port(&status_byte, 1);
						notemidi_write_to_port(&second_byte, 1);
						copy_from_user(&third_byte, &data[scan + 5], 1);
						notemidi_write_to_port(&third_byte, 1);
						//printk(KERN_INFO "NOTEMIDI: Echoing note %x %x %x\n", status_byte, second_byte, third_byte);
					}

					if (retry)
						scan += 8;
					break;
			}
	}

	/***
	 if (scan)
		printk(KERN_INFO "NOTEMIDI: %d bytes read\n", scan);
	***/

	return scan;
}

static int notemidi_file_ioctl (struct inode *inode, struct file *file, unsigned int command, unsigned long data)
{
	/* printk(KERN_INFO "NOTEMIDI: ioctl %x\n", command); */

	if (command == SNDCTL_SEQ_NRSYNTHS)
	{
		int *count = (int*) data;
		put_user(1, count);
	}

	else if (command == SNDCTL_SEQ_NRMIDIS)
	{
		int *count = (int*) data;
		put_user(1, count);
	}

	else if (command == SNDCTL_SYNTH_INFO)
	{
		struct synth_info info;
		struct synth_info *user_info = (struct synth_info *) data;

		memset(&info, 0, sizeof(info));
		strcpy(info.name, "NOTEMIDI");
		info.synth_type = SYNTH_TYPE_MIDI;
		//info.synth_subtype = MIDI_TYPE_MPU401;
		info.nr_voices = 64; /* arbitrarily chosen */
		copy_to_user(user_info, &info, sizeof(info));
	}

	else if (command == SNDCTL_MIDI_INFO)
	{
		struct midi_info info;
		struct midi_info *user_info = (struct midi_info *) data;

		memset(&info, 0, sizeof(info));
		strcpy(info.name, "NOTEMIDI");
		copy_to_user(user_info, &info, sizeof(info));
	}

	else if (command == SNDCTL_SEQ_RESET)
	{
		/* doesn't actually do anything - but acknowledge so there are no problems */
		notemidi_open_time = jiffies;
		notemidi_timer_start = 0;
	}

	else if (command == SNDCTL_SEQ_SYNC)
	{
		/* doesn't actually do anything - but acknowledge so there are no problems */
		// notemidi_open_time = jiffies;
	}

	else if (command == SNDCTL_SEQ_CTRLRATE)
	{
		int *timer_resolution = (int*) data;
		put_user(HZ, timer_resolution);
	}

	else if (command == SNDCTL_SEQ_OUTOFBAND)
	{
		/* doesn't actually do anything - but acknowledge so there are no problems */
	}

	else if (command == SNDCTL_TMR_TIMEBASE)
	{
		int *new_base = (int*) data;
		get_user(notemidi_timer_base, new_base);
		/* printk(KERN_INFO "NOTEMIDI: New base %d\n", notemidi_timer_base); */
	}

	else if (command == SNDCTL_TMR_TEMPO)
	{
		int *new_tempo = (int*) data;
		get_user(notemidi_timer_tempo, new_tempo);
		/* printk(KERN_INFO "NOTEMIDI: New tempo %d\n", notemidi_timer_base); */
	}

	else if (command == SNDCTL_SYNTH_MEMAVL)
	{
		/* doesn't actually do anything - but acknowledge so there are no problems */
		int *free_memory = (int*) data;
		put_user(4096, free_memory);
		printk(KERN_INFO "NOTEMIDI: Memavl request\n");
	}

	else
	{
		printk(KERN_INFO "NOTEMIDI: undefined ioctl %x\n", command);
        return -1;
	}

	return 0;
}

struct file_operations notemidi_file_operations =
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
    NULL, /* struct module *owner; */
#endif
	NULL, /* loff_t (*llseek) (struct file *, loff_t, int); */
	notemidi_file_read, /* ssize_t (*read) (struct file *, char *, size_t, loff_t *); */
	notemidi_file_write, /* ssize_t (*write) (struct file *, const char *, size_t, loff_t *); */
	NULL, /* int (*readdir) (struct file *, void *, filldir_t); */
	NULL, /* unsigned int (*poll) (struct file *, struct poll_table_struct *); */
	notemidi_file_ioctl, /* int (*ioctl) (struct inode *, struct file *, unsigned int, unsigned long); */
	NULL, /* int (*mmap) (struct file *, struct vm_area_struct *); */
	notemidi_file_open, /* int (*open) (struct inode *, struct file *); */
	NULL, /* int (*flush) (struct file *); */
	notemidi_file_release, /* int (*release) (struct inode *, struct file *); */
	NULL, /* int (*fsync) (struct file *, struct dentry *); */
	NULL, /* int (*fasync) (int, struct file *, int); */
	NULL, /* int (*lock) (struct file *, int, struct file_lock *); */
	NULL, /* ssize_t (*readv) (struct file *, const struct iovec *, unsigned long, loff_t *); */
	NULL, /* ssize_t (*writev) (struct file *, const struct iovec *, unsigned long, loff_t *); */
/*
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	NULL,  ssize_t (*writepage) (struct file *, struct page *, int, size_t, loff_t *, int);
#endif
*/
};


/************************ Module entry/exit points ********************************/

int init_module(void)
{
	int registration_status = 0;

	notemidi_open_time = 0;
	notemidi_timer_start = 0;

	notemidi_outbuffer_pointer = 0;
	notemidi_outbuffer_length = 0;

	init_timer(&notemidi_buffer_timer);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
	init_waitqueue(&notemidi_buffer_wait_queue);
#else
	init_waitqueue_head(&notemidi_buffer_wait_queue);
#endif

	if ((registration_status = register_chrdev(NOTEMIDI_MAJOR_NUMBER, "NOTEMIDI", &notemidi_file_operations)) < 0)
	{
		printk(KERN_ERR "NOTEMIDI: register_chrdev() failed code %d\n", registration_status);
		return -1;
	}

	else if ((notemidi_serial_port_tty = notemidi_open_serial_port(NOTEMIDI_SERIAL_PORT_DEVICE_NAME)) == NULL)
	{
		unregister_chrdev(NOTEMIDI_MAJOR_NUMBER, "NOTEMIDI");
		return -1;
	}

	printk(KERN_INFO "NOTEMIDI v2001.11.28 loaded -------\n");

	return 0;
}

void cleanup_module(void)
{
	if (waitqueue_active(&notemidi_buffer_wait_queue))
		wake_up_interruptible(&notemidi_buffer_wait_queue);

	del_timer(&notemidi_buffer_timer);

	if (!IS_ERR(notemidi_serial_port_file))
		filp_close(notemidi_serial_port_file, 0);

	unregister_chrdev(NOTEMIDI_MAJOR_NUMBER, "NOTEMIDI");

	notemidi_serial_port_file = 0;
	notemidi_serial_port_tty = 0;

	printk(KERN_INFO "NOTEMIDI unloaded -------------------\n");
}


/*
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
module_init(init_module);
module_exit(cleanup_module);
#endif
*/

