/***************************************************************************
           NOTEMIDIPLAY - simple command line type 1 MIDI file player
                            -------------------
    begin                : June 24, 2001
    copyright            : (C) 2001 by Michael Minn
    email                : michael@michaelminn.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

***************************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>

class midi_file_header
{
	public:
	char chunk_type[4]; // "MThd"

	unsigned char length[4];
	unsigned char format[2];
	unsigned char tracks[2];
	unsigned char division[2];
	char track_type[4]; // "MTrk"
	unsigned char track_length[4];
};

int read_delta_time(int file)
{
	int delta_time = 0;
	unsigned char byte;

	while (read(file, &byte, 1) == 1)
	{
		delta_time <<= 7;
		delta_time += ((int)byte & 0x7F);
		if (!(byte & 0x80)) // bit 7 is low in last byte of delta time
			return delta_time;
	}

	return -1;
}

#define MIDI_NOTE_OFF 0x80
#define MIDI_NOTE_ON  0x90
#define MIDI_POLYPHONIC_KEY_PRESSURE 0xA0
#define MIDI_CONTROL_CHANGE 0xB0
#define MIDI_PROGRAM_CHANGE 0xC0
#define MIDI_AFTERTOUCH 0xD0
#define MIDI_PITCH_BEND 0xE0
#define MIDI_SYSTEM_EXCLUSIVE 0xF0
#define MIDI_TIME_CODE_QUARTER_FRAME 0xF1
#define MIDI_SONG_POSITION 0xF2
#define MIDI_SONG_SELECT 0xF3
#define MIDI_TUNE_REQUEST 0xF6
#define MIDI_END_OF_EXCLUSIVE 0xF7
#define MIDI_TIMING_CLOCK 0xF8
#define MIDI_START 0xFA
#define MIDI_CONTINUE 0xFB
#define MIDI_STOP 0xFC
#define MIDI_ACTIVE_SENSING 0xFE
#define MIDI_META_EVENT 0xFF


int main (int argc, char **argv)
{
	printf ("NOTEMIDIPLAY v2001.07.30 by Michael Minn\n");
	if ((argc < 2) || (argc > 3))
    {
    	printf("SYNTAX: notemidiplay <midi file name> [<audio device name>]\n");
		printf ("Plays type 0 MIDI files to a Linux MIDI device (default /dev/midi03)\n");
		return -1;
	}

	char *output_device_name = "/dev/midi03";
	if (argc == 3)
		output_device_name = argv[2];

    int input_file = open (argv[1], O_RDONLY);
    if (input_file == -1)
    {
		printf("Failure opening input file %s: %s", argv[1], strerror(errno));
        return -1;
	}

    int device_file = open (output_device_name, O_WRONLY);
    if (device_file == -1)
    {
		printf("Failure opening output device %s: %s", output_device_name, strerror(errno));
        return -1;
    }

	midi_file_header header;
	if (read(input_file, &header, sizeof(header)) != sizeof(header))
    {
		perror("Failure reading file header");
        return -1;
    }

	int division = ((int)header.division[0] << 8) + (int)header.division[1];
	int tracks = ((int)header.tracks[0] << 8) + (int)header.tracks[1];
	int format = ((int)header.format[0] << 8) + (int)header.format[1];
	int tempo = 120;

	if ((format != 0) || (tracks != 1))
	{
		printf("Only single track type 0 format files supported\n");
		printf("This file is format type %d (%d tracks)\n", format, tracks);
		return -1;
	}

	int delta_time = 0;
	unsigned char running_status = 0;
	unsigned char message[256];

	struct timezone time_zone;
	struct timeval start_timeval;
	gettimeofday(&start_timeval, &time_zone);
	int needed_microseconds = 0;

	while ((delta_time = read_delta_time(input_file)) >= 0)
	{
		needed_microseconds += (60000000 / division * delta_time  / tempo);

		struct timeval current_timeval;
		gettimeofday(&current_timeval, &time_zone);
		int actual_microseconds = ((current_timeval.tv_sec - start_timeval.tv_sec) * 1000000) + (current_timeval.tv_usec - start_timeval.tv_usec);
		if (needed_microseconds > actual_microseconds)
			usleep(needed_microseconds - actual_microseconds);

		int bytes_read = 1;
		read(input_file, message, 1);

		if (!(message[0] & 0x80))
		{
			message[1] = message[0];
			message[0] = running_status;
			bytes_read = 2;
		}
		running_status = message[0];

		if ((running_status >= MIDI_NOTE_OFF) && (running_status < MIDI_PROGRAM_CHANGE))
		{
			read(input_file, &message[bytes_read], 3 - bytes_read);
			write(device_file, message, 3);
		}

		else if ((running_status >= MIDI_PROGRAM_CHANGE) && (running_status < MIDI_PITCH_BEND))
		{
			read(input_file, &message[bytes_read], 2 - bytes_read);
			write(device_file, message, 2);
		}

		else if ((running_status >= MIDI_PITCH_BEND) && (running_status < MIDI_SYSTEM_EXCLUSIVE))
		{
			read(input_file, &message[bytes_read], 3 - bytes_read);
			write(device_file, message, 3);
		}

		else if ((running_status == MIDI_TIME_CODE_QUARTER_FRAME) || (running_status == MIDI_SONG_SELECT))
		{
			read(input_file, &message[bytes_read], 2 - bytes_read);
			write(device_file, message, 2);
		}

		else if (running_status == MIDI_SONG_POSITION)
		{
			read(input_file, &message[bytes_read], 3 - bytes_read);
			write(device_file, message, 3);
		}

		else if ((running_status >= MIDI_TUNE_REQUEST) && (running_status <= MIDI_ACTIVE_SENSING))
		{
			read(input_file, &message[bytes_read], 2 - bytes_read);
			write(device_file, message, 2);
		}

		else if (running_status == MIDI_META_EVENT)
		{
			read(input_file, &message[bytes_read], 3 - bytes_read);
			read(input_file, &message[3], (int)message[2]);

			if (message[1] == 0x51)
			{
				int microseconds_per_quarter = ((int)message[3] << 16) + ((int)message[4] << 8) + ((int)message[5]);
				tempo = 60000000 / microseconds_per_quarter;
				printf("New Tempo %d\n", tempo);
			}
			//else if (message[1] == 0x05) // lyric
			//	printf("%*s\n", (int)message[2], &message[3]);
		}

		else // system exclusive
		{
			while (message[bytes_read - 1] != MIDI_END_OF_EXCLUSIVE)
				if (read(input_file, &message[bytes_read], 1) < 1)
					return printf("Premature end of sysex message\n");
				else if (bytes_read >= 255)
					return printf("Sysex message too long\n");
				else
					++bytes_read;

			write(device_file, message, bytes_read);
		}
	}

	close(device_file);
    close(input_file);

    return 0;
}
